/*
 *   Copyright (C) 2002,2003 by Jonathan Naylor G4KLX/HB9DRD
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "FSK441App.h"

#include <wx/config.h>
#include <wx/ffile.h>

IMPLEMENT_APP(CFSK441App)

#include "common/Exception.h"
#include "common/Log.h"

#include "fsk441/FSK441Defs.h"

CFSK441App::CFSK441App() :
wxApp(),
m_frame(NULL),
m_controller(NULL),
m_messages(NULL),
m_file(NULL),
m_saving(false),
m_recording(false),
m_txEnable(DEFAULT_TX_ENABLED)
{
}

CFSK441App::~CFSK441App()
{
}

bool CFSK441App::OnInit()
{
	SetVendorName(FSK441_VENDOR_NAME);
	SetAppName(FSK441_APPL_NAME);

	wxConfigBase* config = wxConfigBase::Get();
	config->SetRecordDefaults();

	wxLog* log = new CLog(FSK441_LOGFILE_NAME);
	wxLog::SetActiveTarget(log);

	createController();
	::wxSleep(1);

	createFileReader();

	readRegistry();

	m_frame = new CFSK441Frame();
	m_frame->Show();

	SetTopWindow(m_frame);

	return true;
}

int CFSK441App::OnExit()
{
	wxASSERT(m_controller != NULL);
	wxASSERT(m_file != NULL);

	m_controller->Delete();
	m_file->Delete();

	delete wxConfigBase::Set(NULL);

	return 0;
}

#ifdef __WXDEBUG__
void CFSK441App::OnAssert(const wxChar* file, int line, const wxChar* cond, const wxChar* msg)
{
	::wxLogFatalError(wxT("Assertion failed on line %d in file %s: %s %s"), line, file, cond, msg);
}
#endif

void CFSK441App::receiveCardMessage(CFSK441Message* message) const
{
	wxASSERT(m_frame != NULL);
	wxASSERT(message != NULL);

	m_frame->showCardMessage(message);
}

void CFSK441App::receiveFileMessage(CFSK441Message* message) const
{
	wxASSERT(m_frame != NULL);
	wxASSERT(message != NULL);

	m_frame->showFileMessage(message);
}

void CFSK441App::logMessage(const CFSK441Message& message) const
{
	if (m_saving) {
		wxString  id = message.getId();
		double   tim = message.getTime();
		int      len = message.getLength();
		int strength = message.getStrength();

		wxString text = message.getText();

		wxFFile* file = new wxFFile(FSK441_MESSAGES_NAME, "a");

		if (!file->IsOpened()) {
			wxString text;
			text.Printf(wxT("Cannot open save text file: %s"), FSK441_MESSAGES_NAME.c_str());
			error(text);

			delete file;
			return;
		}

		wxString line;
		line.Printf(wxT("%s  %.1f  %d  %d  %s\n"), id.c_str(), tim, len, strength, text.c_str());

		file->Write(line);
		file->Close();

		delete file;
	}
}

void CFSK441App::showCardLevels(CFSK441Levels* levels) const
{
	wxASSERT(levels != NULL);
	wxASSERT(m_frame != NULL);

	m_frame->showCardLevels(levels);
}

void CFSK441App::showFileLevels(CFSK441Levels* levels) const
{
	wxASSERT(levels != NULL);
	wxASSERT(m_frame != NULL);

	m_frame->showFileLevels(levels);
}

void CFSK441App::showCardSpectrum(CFSK441Spectrum* spectrum) const
{
	wxASSERT(spectrum != NULL);
	wxASSERT(m_frame != NULL);

	m_frame->showCardSpectrum(spectrum);
}

void CFSK441App::showFileSpectrum(CFSK441Spectrum* spectrum) const
{
	wxASSERT(spectrum != NULL);
	wxASSERT(m_frame != NULL);

	m_frame->showFileSpectrum(spectrum);
}

void CFSK441App::error(const wxString& text) const
{
	wxASSERT(m_frame != NULL);

	m_frame->error(text);
}

void CFSK441App::getPersonal(wxString& callsign, wxString& locator) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString callsignKey = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_CALLSIGN;
	wxString locatorKey  = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_LOCATOR;

	profile->Read(callsignKey, &callsign, DEFAULT_CALLSIGN);
	profile->Read(locatorKey,  &locator,  DEFAULT_LOCATOR);
}

void CFSK441App::setPersonal(const wxString& callsign, const wxString& locator) const
{
	CText::setLocalCallsign(callsign);
	CText::setLocalLocator(locator);

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString callsignKey = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_CALLSIGN;
	wxString locatorKey  = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_LOCATOR;

	profile->Write(callsignKey, callsign);
	profile->Write(locatorKey,  locator);
	profile->Flush();
}

void CFSK441App::setRemote(const wxString& callsign, const wxString& report)
{
	if (callsign.IsEmpty())
		CText::setRemoteCallsign(wxEmptyString);
	else
		CText::setRemoteCallsign(callsign);

	if (report.IsEmpty())
		CText::setLocalReport(wxEmptyString);
	else
		CText::setLocalReport(report);
}

bool CFSK441App::getTXFirst() const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_TRANSMIT + wxT("/") + KEY_FIRST;

	bool txFirst;
	profile->Read(key, &txFirst, DEFAULT_TX_FIRST);

	return txFirst;
}

void CFSK441App::setTXFirst(bool txFirst) const
{
	wxASSERT(m_controller != NULL);

	m_controller->setTXFirst(txFirst);

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_TRANSMIT + wxT("/") + KEY_FIRST;

	profile->Write(key, txFirst);
	profile->Flush();
}

bool CFSK441App::getTXEnable() const
{
	return m_txEnable;
}

void CFSK441App::setTXEnable(bool txEnable)
{
	wxASSERT(m_controller != NULL);

	m_controller->setTXEnable(txEnable);

	m_txEnable = txEnable;
}

bool CFSK441App::getSingleTone() const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_PROTOCOL + wxT("/") + KEY_SINGLE_TONE;

	bool singleTone;
	profile->Read(key, &singleTone, DEFAULT_SINGLE_TONE);

	return singleTone;
}

void CFSK441App::setSingleTone(bool singleTone) const
{
	wxASSERT(m_controller != NULL);

	m_controller->setSingleTone(singleTone);

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_PROTOCOL + wxT("/") + KEY_SINGLE_TONE;

	profile->Write(key, singleTone);
	profile->Flush();
}

int CFSK441App::getMinLength() const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_RECEIVE + wxT("/") + KEY_MIN_LENGTH;

	long minLength;
	profile->Read(key, &minLength, DEFAULT_MIN_LENGTH);

	return int(minLength);
}

void CFSK441App::setMinLength(int minLength) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_RECEIVE + wxT("/") + KEY_MIN_LENGTH;

	profile->Write(key, minLength);
	profile->Flush();
}

int CFSK441App::getMinStrength() const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_RECEIVE + wxT("/") + KEY_MIN_STRENGTH;

	long minStrength;
	profile->Read(key, &minStrength, DEFAULT_MIN_STRENGTH);

	return int(minStrength);
}

void CFSK441App::setMinStrength(int minStrength) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_RECEIVE + wxT("/") + KEY_MIN_STRENGTH;

	profile->Write(key, minStrength);
	profile->Flush();
}

wxString CFSK441App::getReplacedMessage(int n) const
{
	wxASSERT(n >= 0 && n < FSK441_MAX_MESSAGES);
	wxASSERT(m_messages[n] != NULL);

	return m_messages[n]->getReplacedText();
}

wxString CFSK441App::getOriginalMessage(int n) const
{
	wxASSERT(n >= 0 && n < FSK441_MAX_MESSAGES);
	wxASSERT(m_messages[n] != NULL);

	return m_messages[n]->getOriginalText();
}

void CFSK441App::getMessage(int n, wxString& message) const
{
	wxASSERT(n >= 0 && n < FSK441_MAX_MESSAGES);

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString number;
	number.Printf(wxT("%d"), n);

	wxString key = wxT("/") + SECTION_MESSAGES + wxT("/") + KEY_MESSAGE + number;
	wxString defMessage;

	switch (n) {
		case 0:
			defMessage = DEFAULT_MESSAGE_0;
			break;
		case 1:
			defMessage = DEFAULT_MESSAGE_1;
			break;
		case 2:
			defMessage = DEFAULT_MESSAGE_2;
			break;
		case 3:
			defMessage = DEFAULT_MESSAGE_3;
			break;
		case 4:
			defMessage = DEFAULT_MESSAGE_4;
			break;
		case 5:
			defMessage = DEFAULT_MESSAGE_5;
			break;
		default:
			defMessage = wxEmptyString;
			break;
	}

	profile->Read(key, &message, defMessage);
}

void CFSK441App::setMessage(int n, const wxString& message)
{
	wxASSERT(n >= 0 && n < FSK441_MAX_MESSAGES);

	if (m_messages[n] != NULL)
		delete m_messages[n];

	m_messages[n] = new CText(message);

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString number;
	number.Printf(wxT("%d"), n);

	wxString key = wxT("/") + SECTION_MESSAGES + wxT("/") + KEY_MESSAGE + number;

	profile->Write(key, message);
	profile->Flush();
}

void CFSK441App::getPTT(wxString& type, wxString& device) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString typeKey   = wxT("/") + SECTION_PTT + wxT("/") + KEY_TYPE;
	wxString deviceKey = wxT("/") + SECTION_PTT + wxT("/") + KEY_DEVICE;

	profile->Read(typeKey,   &type,   DEFAULT_PTT_TYPE);
	profile->Read(deviceKey, &device, DEFAULT_PTT_DEVICE);
}

void CFSK441App::setPTT(const wxString& type, const wxString& device) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString typeKey   = wxT("/") + SECTION_PTT + wxT("/") + KEY_TYPE;
	wxString deviceKey = wxT("/") + SECTION_PTT + wxT("/") + KEY_DEVICE;

	profile->Write(typeKey,   type);
	profile->Write(deviceKey, device);
	profile->Flush();
}

void CFSK441App::getSound(wxString& type, wxString& fileName, int& bits) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString typeKey     = wxT("/") + SECTION_SOUND + wxT("/") + KEY_TYPE;
	wxString fileNameKey = wxT("/") + SECTION_SOUND + wxT("/") + KEY_FILENAME;
	wxString bitsKey     = wxT("/") + SECTION_SOUND + wxT("/") + KEY_BITS;

	long longBits;

	profile->Read(typeKey,     &type,     DEFAULT_SOUND_TYPE);
	profile->Read(fileNameKey, &fileName, DEFAULT_SOUND_FILENAME);
	profile->Read(bitsKey,     &longBits, DEFAULT_SOUND_BITS);

	bits = int(longBits);
}

void CFSK441App::setSound(const wxString& type, const wxString& fileName, int bits) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString typeKey     = wxT("/") + SECTION_SOUND + wxT("/") + KEY_TYPE;
	wxString fileNameKey = wxT("/") + SECTION_SOUND + wxT("/") + KEY_FILENAME;
	wxString bitsKey     = wxT("/") + SECTION_SOUND + wxT("/") + KEY_BITS;

	long longBits = long(bits);

	profile->Write(typeKey,     type);
	profile->Write(fileNameKey, fileName);
	profile->Write(bitsKey,     longBits);
	profile->Flush();
}

void CFSK441App::setSaving(bool saving)
{
	m_saving = saving;
}

void CFSK441App::deleteMessages() const
{
	::wxRemoveFile(FSK441_MESSAGES_NAME);
}

void CFSK441App::setRecording(bool recording)
{
	wxASSERT(m_controller != NULL);

	m_recording = recording;

	m_controller->setRXRecording(m_recording);
}

void CFSK441App::sendMessage(const wxString& text)
{
	wxASSERT(m_controller != NULL);

	m_controller->sendMessage(text);
}

void CFSK441App::readFile(const wxString& fileName)
{
	wxASSERT(m_file != NULL);

	if (!m_file->IsPaused()) {
		error(wxT("Another WAV file is currently being processed"));
		return;
	}

	try {
		bool singleTone = getSingleTone();

		m_file->setFileName(fileName);
		m_file->setSingleTone(singleTone);

		m_file->Resume();
		m_file->Pause();
	}
	catch (CException& ex) {
		error(ex.getMessage());
		return;
	}
}

void CFSK441App::createController()
{
	try {
		wxString type, fileName;
		int bits;

		getSound(type, fileName, bits);
		CSoundDev* soundDev = CSoundDev::createSoundDev(type, fileName, FSK441_SAMPLE_RATE, bits);

		getPTT(type, fileName);
		CPTTPort* pttPort = CPTTPort::createPTTPort(type, fileName);

		m_controller = new CFSK441Controller(soundDev, pttPort);
		m_controller->Create();
		m_controller->Run();
	}
	catch (CException& ex) {
		::fprintf(stderr, "FSK441: an error has occured when creating the sound device or the PTT port: %s\n", ex.getMessage().c_str());
		throw;
	}
}

void CFSK441App::createFileReader()
{
	m_file = new CFSK441FileReceive();

	m_file->Create();
	m_file->Run();
	m_file->Pause();
}

void CFSK441App::readRegistry()
{
	wxASSERT(m_controller != NULL);

	wxString callsign, locator;
	getPersonal(callsign, locator);
	CText::setLocalCallsign(callsign);
	CText::setLocalLocator(locator);

	bool txEnable = getTXEnable();
	m_controller->setTXEnable(txEnable);

	bool txFirst = getTXFirst();
	m_controller->setTXFirst(txFirst);

	bool singleTone = getSingleTone();
	m_controller->setSingleTone(singleTone);

	m_messages = new CText*[FSK441_MAX_MESSAGES];

	for (int i = 0; i < FSK441_MAX_MESSAGES; i++) {
		wxString message;
		getMessage(i, message);
		m_messages[i] = new CText(message);
	}

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	profile->Flush();
}
