/*
 *   Copyright (C) 2002,2003 by Jonathan Naylor G4KLX/HB9DRD
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "FSK441Display.h"
#include "FSK441App.h"

#include "fsk441/FSK441Defs.h"

const int BORDER_SIZE = 5;

enum {
	List_Messages        = 150,
	Menu_Messages_Clear  = 250
};

BEGIN_EVENT_TABLE(CFSK441Display, wxPanel)
	EVT_MENU(Menu_Messages_Clear, CFSK441Display::onClearMessages)
	EVT_LIST_ITEM_SELECTED(List_Messages, CFSK441Display::onMessageSelected)
	EVT_LIST_ITEM_RIGHT_CLICK(List_Messages, CFSK441Display::onMessageClick)
END_EVENT_TABLE()

CFSK441Display::CFSK441Display(wxWindow* parent, int id) :
wxPanel(parent, id),
m_levelGraph(NULL),
m_spectrumGraph(NULL),
m_messageListCtrl(NULL),
m_messageMenu(NULL),
m_listCount(0),
m_selected(-1),
m_minLength(40),
m_minStrength(3),
m_messageList(NULL)
{
	wxASSERT(parent != NULL);

	m_messageMenu = createMessageMenu();

	m_messageList = new CFSK441MessageList(wxKEY_NONE);
	m_messageList->DeleteContents(true);

	wxBoxSizer* sizer = new wxBoxSizer(wxVERTICAL);

	wxBoxSizer* graphSizer = new wxBoxSizer(wxHORIZONTAL);

	m_levelGraph = new CFSK441LevelGraph(this, -1, wxDefaultPosition, wxSize(FSK441_LEVEL_WIDTH, FSK441_GRAPH_HEIGHT));
	graphSizer->Add(m_levelGraph, 0, wxLEFT | wxBOTTOM | wxRIGHT, BORDER_SIZE);

	m_spectrumGraph = new CFSK441SpectrumGraph(this, -1, wxDefaultPosition, wxSize(FSK441_SPECTRUM_WIDTH, FSK441_GRAPH_HEIGHT));
	graphSizer->Add(m_spectrumGraph, 0, wxALL, 0);

	sizer->Add(graphSizer, 0, wxTOP, BORDER_SIZE);

	m_messageListCtrl = new wxListCtrl(this, List_Messages, wxDefaultPosition, wxSize(FSK441_MESSAGES_WIDTH, FSK441_MESSAGES_HEIGHT));
	m_messageListCtrl->SetSingleStyle(wxLC_REPORT);
	m_messageListCtrl->InsertColumn(0, wxT("Id"));
	m_messageListCtrl->SetColumnWidth(0, 75);
	m_messageListCtrl->InsertColumn(1, wxT("Time"));
	m_messageListCtrl->SetColumnWidth(1, 50);
	m_messageListCtrl->InsertColumn(2, wxT("Length"));
	m_messageListCtrl->SetColumnWidth(2, 50);
	m_messageListCtrl->InsertColumn(3, wxT("Strength"));
	m_messageListCtrl->SetColumnWidth(3, 60);
	m_messageListCtrl->InsertColumn(4, wxT("Text"));
	m_messageListCtrl->SetColumnWidth(4, 345);

	sizer->Add(m_messageListCtrl, 0, wxALL, 0);

	SetAutoLayout(true);
	sizer->Fit(this);
	sizer->SetSizeHints(this);
	SetSizer(sizer);
}

CFSK441Display::~CFSK441Display()
{
	delete m_messageList;
}

wxMenu* CFSK441Display::createMessageMenu()
{
	wxMenu* menu = new wxMenu();
	menu->Append(Menu_Messages_Clear, wxT("Clear"));

	return menu;
}

void CFSK441Display::onMessageSelected(const wxListEvent& event)
{
	m_selected = event.GetIndex();
}

void CFSK441Display::onMessageClick(const wxListEvent& event)
{
	wxASSERT(m_messageListCtrl != NULL);

	if (m_selected != -1) {
		wxPoint point = event.GetPoint();
		m_messageListCtrl->PopupMenu(m_messageMenu, point);
	}
}

void CFSK441Display::onClearMessages(const wxCommandEvent& event)
{
	wxASSERT(m_messageListCtrl != NULL);
	wxASSERT(m_messageList != NULL);

	m_messageListCtrl->DeleteAllItems();
	m_messageList->Clear();

	m_listCount = 0;
	m_selected  = -1;
}

void CFSK441Display::showMessage(CFSK441Message* message)
{
	wxASSERT(m_spectrumGraph != NULL);
	wxASSERT(m_messageListCtrl != NULL);
	wxASSERT(message != NULL);
	wxASSERT(m_listCount >= 0);

	if (message->getLength() < m_minLength ||
	    message->getStrength() < m_minStrength)
		return;

	wxString id = message->getId();

	int pos = id.Find(wxT('_'), true);
	if (pos != -1) {
		wxString tmp = id.Mid(pos + 1);
		id = tmp;
	}

	m_messageListCtrl->InsertItem(0, id, 0);

	wxString text = message->getText();

	wxString buffer;
	buffer.Printf(wxT("%.1f"), message->getTime());
	m_messageListCtrl->SetItem(0, 1, buffer);

	buffer.Printf(wxT("%d"), message->getLength());
	m_messageListCtrl->SetItem(0, 2, buffer);

	buffer.Printf(wxT("%d"), message->getStrength());
	m_messageListCtrl->SetItem(0, 3, buffer);
	m_messageListCtrl->SetItem(0, 4, text);

	::wxLogMessage(wxT("Id:%s Start:%.1fs Length:%dms Strength:%ddB Text:%s\n"), id.c_str(), message->getTime(), message->getLength(), message->getStrength(), text.c_str());
	::wxGetApp().logMessage(*message);

	m_listCount++;

	pos = 0;
	m_messageList->Insert(pos, message);

	// The list is moving down and the selected message has moved down
	// by one
	if (m_selected != -1)
		m_selected++;
}

void CFSK441Display::showLevels(CFSK441Levels* levels) const
{
	wxASSERT(levels != NULL);
	wxASSERT(m_levelGraph != NULL);

	m_levelGraph->addData(levels);
}

void CFSK441Display::showSpectrum(CFSK441Spectrum* spectrum) const
{
	wxASSERT(spectrum != NULL);
	wxASSERT(m_levelGraph != NULL);

	m_spectrumGraph->addData(spectrum);
}

void CFSK441Display::setMinLength(int length)
{
	wxASSERT(length >= 20);

	m_minLength = length;
}

void CFSK441Display::setMinStrength(int strength)
{
	wxASSERT(strength >= 1);

	m_minStrength = strength;
}
